==> mule-tooling-bootstrap-1.0.0-sources.jar: org/mule/tooling/client/bootstrap/internal/DefaultToolingRuntimeClientBootstrap.java <==
/*
 * Copyright (c) MuleSoft, Inc.  All rights reserved.  http://www.mulesoft.com
 * The software in this package is published under the terms of the CPAL v1.0
 * license, a copy of which has been included with this distribution in the
 * LICENSE.txt file.
 */

package org.mule.tooling.client.bootstrap.internal;

import static com.google.common.io.Files.createTempDir;
import static com.google.common.io.Files.getFileExtension;
import static java.lang.Integer.max;
import static java.lang.Runtime.getRuntime;
import static java.lang.String.format;
import static java.lang.System.clearProperty;
import static java.lang.System.identityHashCode;
import static java.lang.System.setProperty;
import static java.util.Objects.requireNonNull;
import static java.util.concurrent.Executors.newFixedThreadPool;
import static java.util.stream.Collectors.toList;
import static org.apache.commons.io.FileUtils.copyInputStreamToFile;
import static org.apache.commons.io.FileUtils.deleteQuietly;
import static org.apache.commons.lang3.ArrayUtils.addAll;
import static org.mule.tooling.client.api.descriptors.ArtifactDescriptor.newBuilder;
import static org.mule.tooling.client.bootstrap.internal.VersionResolver.mapToolingVersion;
import org.mule.maven.client.api.MavenClient;
import org.mule.maven.client.api.MavenClientProvider;
import org.mule.maven.client.api.model.BundleDependency;
import org.mule.maven.client.api.model.BundleDescriptor;
import org.mule.runtime.module.embedded.internal.classloading.FilteringClassLoader;
import org.mule.runtime.module.embedded.internal.classloading.JdkOnlyClassLoaderFactory;
import org.mule.tooling.client.api.ToolingRuntimeClient;
import org.mule.tooling.client.api.ToolingRuntimeClientBuilderFactory;
import org.mule.tooling.client.api.descriptors.ArtifactDescriptor;
import org.mule.tooling.client.api.exception.ToolingException;
import org.mule.tooling.client.bootstrap.api.ToolingRuntimeClientBootstrap;
import org.mule.tooling.client.bootstrap.api.ToolingRuntimeClientBootstrapConfiguration;
import org.mule.tooling.client.bootstrap.internal.classloader.ToolingClassLoader;
import org.mule.tooling.client.bootstrap.internal.wrapper.ToolingRuntimeClientBuilderFactoryWrapper;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ExecutorService;

import org.apache.commons.lang3.ArrayUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Bootstrap for {@link ToolingRuntimeClient}.
 *
 * @since 1.0
 */
public class DefaultToolingRuntimeClientBootstrap implements ToolingRuntimeClientBootstrap {

  private static final String JAR = "jar";
  private static final int AVAILABLE_PROCESSORS = getRuntime().availableProcessors();

  private static final String LOG4J2_DISABLE_JMX = "log4j2.disable.jmx";
  private static final String LOG4J2_LOGGER_CONTEXT_FACTORY = "log4j2.loggerContextFactory";

  private static Logger LOGGER = LoggerFactory.getLogger(DefaultToolingRuntimeClientBootstrap.class);

  private static final String ORG_MULE_TOOLING = "org.mule.tooling";
  private static final String MULE_RUNTIME_TOOLING_CLIENT = "mule-runtime-tooling-client";

  private final ToolingRuntimeClientBootstrapConfiguration configuration;

  private final MavenClient mavenClient;
  private URLClassLoader toolingClassLoader;
  private ToolingRuntimeClientBuilderFactoryWrapper toolingRuntimeClientBuilderFactory;

  private File confDir = null;

  private ExecutorService executorService;
  private int threadNumber = 1;

  /**
   * Creates an instance of the Bootstrap.
   *
   * @param configuration {@link ToolingRuntimeClientBootstrapConfiguration} that defines the configuration for this bootstrap. Not null.
   */
  public DefaultToolingRuntimeClientBootstrap(ToolingRuntimeClientBootstrapConfiguration configuration) {
    requireNonNull(configuration, "configuration cannot be null");
    this.configuration = configuration;

    LOGGER.info("Creating a Tooling Bootstrap for muleVersion: {}", configuration.muleVersion());

    // Log4j configuration for Tooling
    String currentLog4jDisableJmx = setProperty(LOG4J2_DISABLE_JMX, "true");
    String currentLog4jLoggerContextFactory =
        setProperty(LOG4J2_LOGGER_CONTEXT_FACTORY, "org.mule.tooling.client.internal.log4j.ToolingLog4jContextFactory");
    try {
      configuration.log4jConfiguration().ifPresent(log4JConfigurationFile -> {
        try (InputStream log4jConfFileInputStream = log4JConfigurationFile.toURL().openStream()) {
          confDir = createTempDir();
          // Use log4j2-test.xml as it will be the first file to be loaded using TCCL
          copyInputStreamToFile(log4jConfFileInputStream, new File(confDir, "log4j2-test.xml"));
        } catch (Exception e) {
          throw new ToolingException("Error while writing log4j configuration file", e);
        }
      });

      this.mavenClient = MavenClientProvider.discoverProvider(DefaultToolingRuntimeClientBootstrap.class.getClassLoader())
          .createMavenClient(configuration.mavenConfiguration());
      this.toolingClassLoader = createClassLoader(configuration.muleVersion());

      int nThreads = max(AVAILABLE_PROCESSORS * 2, 2);
      if (configuration.executorServiceConfiguration().isPresent()) {
        nThreads = configuration.executorServiceConfiguration().get().maxNumberOfThreads();
      }
      executorService = newFixedThreadPool(nThreads,
                                           runnable -> {
                                             Thread thread = new Thread(runnable, format("ToolingClient-%s-%s", threadNumber++,
                                                                                         identityHashCode(this)));
                                             thread.setUncaughtExceptionHandler((t, e) -> LOGGER.error("Internal error", e));
                                             return thread;
                                           });

      this.toolingRuntimeClientBuilderFactory =
          new ToolingRuntimeClientBuilderFactoryWrapper(toolingClassLoader,
                                                        executorService,
                                                        configuration.mavenConfiguration());
    } finally {
      resetSystemProperty(LOG4J2_DISABLE_JMX, currentLog4jDisableJmx);
      resetSystemProperty(LOG4J2_LOGGER_CONTEXT_FACTORY, currentLog4jLoggerContextFactory);
    }
  }

  private static void resetSystemProperty(String propertyName, String oldValue) {
    if (oldValue == null) {
      clearProperty(propertyName);
    } else {
      setProperty(propertyName, oldValue);
    }
  }

  private URLClassLoader createClassLoader(String muleVersion) {
    try {
      if (LOGGER.isDebugEnabled()) {
        LOGGER.debug("Creating URL class loader for muleVersion: {}", muleVersion);
      }

      String toolingVersion = mapToolingVersion(this.getClass().getClassLoader(), muleVersion);

      Set<File> classpath = new LinkedHashSet<>();
      classpath.addAll(resolveDependency(newBuilder()
          .withGroupId(ORG_MULE_TOOLING)
          .withArtifactId(MULE_RUNTIME_TOOLING_CLIENT)
          .withVersion(toolingVersion)
          .build()));

      BundleDependency bundleDependency =
          mavenClient.resolveBundleDescriptor(new BundleDescriptor.Builder().setGroupId(ORG_MULE_TOOLING)
              .setArtifactId(MULE_RUNTIME_TOOLING_CLIENT)
              .setVersion(toolingVersion).setType(JAR).build());

      URL[] urls = new URL[] {};
      urls = addAll(urls, getBundleUrl(bundleDependency));
      urls = addAll(urls, classpath.stream()
          .filter(file -> getFileExtension(file.getAbsolutePath()).equals(JAR))
          .map(file -> toUrl(file))
          .toArray(URL[]::new));

      FilteringClassLoader jdkOnlyClassLoader = JdkOnlyClassLoaderFactory.create();
      if (confDir != null) {
        urls = ArrayUtils.addAll(new URL[] {confDir.toURI().toURL()}, urls);
      }

      URLClassLoader classLoader =
          new ToolingClassLoader(configuration.muleVersion(), identityHashCode(this), urls, jdkOnlyClassLoader);

      if (LOGGER.isDebugEnabled()) {
        LOGGER.debug("Created Tooling class loader: {}", classLoader);
      }

      return classLoader;
    } catch (Exception e) {
      throw new ToolingException("Couldn't create the class loader for the tooling client", e);
    }
  }

  private URL getBundleUrl(BundleDependency bundleDependency) {
    try {
      return bundleDependency.getBundleUri().toURL();
    } catch (MalformedURLException e) {
      throw new ToolingException("Error while getting bundleDependency URL", e);
    }
  }

  private List<File> resolveDependency(ArtifactDescriptor artifactDescriptor) {
    if (LOGGER.isDebugEnabled()) {
      LOGGER.debug("Resolving Tooling Runtime Client implementation libraries from: {}", artifactDescriptor);
    }

    List<BundleDependency> bundleDependencies = mavenClient
        .resolveBundleDescriptorDependencies(false, true,
                                             new BundleDescriptor.Builder()
                                                 .setGroupId(artifactDescriptor.getGroupId())
                                                 .setArtifactId(artifactDescriptor.getArtifactId())
                                                 .setVersion(artifactDescriptor.getVersion())
                                                 .setType(artifactDescriptor.getExtension())
                                                 .setClassifier(artifactDescriptor.getClassifier())
                                                 .build());

    return bundleDependencies.stream()
        .filter(bundleDependency -> bundleDependency.getBundleUri() != null)
        .map(bundleDependency -> new File(getBundleUrl(bundleDependency).getFile()))
        .collect(toList());
  }

  private static URL toUrl(File file) {
    try {
      return file.toURI().toURL();
    } catch (MalformedURLException e) {
      throw new ToolingException("Couldn't get URL for file: " + file, e);
    }
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public String getMuleVersion() {
    return configuration.muleVersion();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public ToolingRuntimeClientBuilderFactory getToolingRuntimeClientBuilderFactory() {
    if (toolingRuntimeClientBuilderFactory == null) {
      throw new IllegalStateException("Cannot be created a ToolingRuntimeClientBuilder once the bootstrap has been disposed");
    }
    return toolingRuntimeClientBuilderFactory;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void dispose() {
    try {
      toolingRuntimeClientBuilderFactory.dispose();
    } finally {
      toolingRuntimeClientBuilderFactory = null;
    }

    executorService.shutdownNow();
    executorService = null;

    if (LOGGER.isDebugEnabled()) {
      LOGGER.debug("Closing Tooling class loader: {}", toolingClassLoader);
    }
    closeClassLoader(toolingClassLoader);
    toolingClassLoader = null;
    deleteQuietly(confDir);
  }

  private void closeClassLoader(URLClassLoader classLoader) {
    try {
      classLoader.close();
    } catch (IOException e) {
      // Do nothing
      if (LOGGER.isDebugEnabled()) {
        LOGGER.debug("Error while closing Tooling class loader", e);
      }
    }
  }

}

==> mule-tooling-bootstrap-1.1.0-sources.jar: org/mule/tooling/client/bootstrap/internal/DefaultToolingRuntimeClientBootstrap.java <==
/*
 * Copyright (c) MuleSoft, Inc.  All rights reserved.  http://www.mulesoft.com
 * The software in this package is published under the terms of the CPAL v1.0
 * license, a copy of which has been included with this distribution in the
 * LICENSE.txt file.
 */

package org.mule.tooling.client.bootstrap.internal;

import static com.google.common.io.Files.createTempDir;
import static com.google.common.io.Files.getFileExtension;
import static java.lang.Integer.max;
import static java.lang.Runtime.getRuntime;
import static java.lang.String.format;
import static java.lang.System.clearProperty;
import static java.lang.System.identityHashCode;
import static java.lang.System.setProperty;
import static java.util.Objects.requireNonNull;
import static java.util.concurrent.Executors.newFixedThreadPool;
import static java.util.stream.Collectors.toList;
import static org.apache.commons.io.FileUtils.copyInputStreamToFile;
import static org.apache.commons.io.FileUtils.deleteQuietly;
import static org.apache.commons.lang3.ArrayUtils.addAll;
import static org.mule.tooling.client.api.descriptors.ArtifactDescriptor.newBuilder;
import org.mule.maven.client.api.MavenClient;
import org.mule.maven.client.api.MavenClientProvider;
import org.mule.maven.client.api.model.BundleDependency;
import org.mule.maven.client.api.model.BundleDescriptor;
import org.mule.runtime.module.embedded.internal.classloading.ClassLoaderFilter;
import org.mule.runtime.module.embedded.internal.classloading.ExportedService;
import org.mule.runtime.module.embedded.internal.classloading.FilteringClassLoader;
import org.mule.runtime.module.embedded.internal.classloading.JreExplorer;
import org.mule.tooling.client.api.ToolingRuntimeClient;
import org.mule.tooling.client.api.ToolingRuntimeClientBuilderFactory;
import org.mule.tooling.client.api.descriptors.ArtifactDescriptor;
import org.mule.tooling.client.api.exception.ToolingException;
import org.mule.tooling.client.bootstrap.api.ToolingRuntimeClientBootstrap;
import org.mule.tooling.client.bootstrap.api.ToolingRuntimeClientBootstrapConfiguration;
import org.mule.tooling.client.bootstrap.internal.classloader.ToolingClassLoader;
import org.mule.tooling.client.bootstrap.internal.wrapper.ToolingRuntimeClientBuilderFactoryWrapper;

import com.google.common.collect.ImmutableSet;

import java.io.File;
import java.io.IOException;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLClassLoader;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ExecutorService;

import org.apache.commons.lang3.ArrayUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Bootstrap for {@link ToolingRuntimeClient}.
 *
 * @since 1.0
 */
public class DefaultToolingRuntimeClientBootstrap implements ToolingRuntimeClientBootstrap {

  private static final String JAR = "jar";
  private static final int AVAILABLE_PROCESSORS = getRuntime().availableProcessors();

  private static final String LOG4J2_DISABLE_JMX = "log4j2.disable.jmx";
  private static final String LOG4J2_LOGGER_CONTEXT_FACTORY = "log4j2.loggerContextFactory";

  private static Logger LOGGER = LoggerFactory.getLogger(DefaultToolingRuntimeClientBootstrap.class);

  protected static final String ORG_MULE_TOOLING = "org.mule.tooling";
  protected static final String MULE_RUNTIME_TOOLING_CLIENT = "mule-runtime-tooling-client";

  private final ToolingRuntimeClientBootstrapConfiguration configuration;

  private final MavenClient mavenClient;
  private URLClassLoader toolingClassLoader;
  private ToolingRuntimeClientBuilderFactoryWrapper toolingRuntimeClientBuilderFactory;

  private File confDir = null;

  private ExecutorService executorService;
  private int threadNumber = 1;

  /**
   * Creates an instance of the Bootstrap.
   *
   * @param configuration {@link ToolingRuntimeClientBootstrapConfiguration} that defines the configuration for this bootstrap. Not null.
   */
  public DefaultToolingRuntimeClientBootstrap(ToolingRuntimeClientBootstrapConfiguration configuration) {
    requireNonNull(configuration, "configuration cannot be null");
    this.configuration = configuration;

    LOGGER.info("Bootstrapping a Tooling Runtime Client version: {} for muleVersion: {}", configuration.toolingVersion(),
                configuration.muleVersion());

    // Log4j configuration for Tooling
    String currentLog4jDisableJmx = setProperty(LOG4J2_DISABLE_JMX, "true");
    String currentLog4jLoggerContextFactory =
        setProperty(LOG4J2_LOGGER_CONTEXT_FACTORY, "org.mule.tooling.client.internal.log4j.ToolingLog4jContextFactory");
    try {
      configuration.log4jConfiguration().ifPresent(log4JConfigurationFile -> {
        try (InputStream log4jConfFileInputStream = log4JConfigurationFile.toURL().openStream()) {
          confDir = createTempDir();
          // Use log4j2-test.xml as it will be the first file to be loaded using TCCL
          copyInputStreamToFile(log4jConfFileInputStream, new File(confDir, "log4j2-test.xml"));
        } catch (Exception e) {
          throw new ToolingException("Error while writing log4j configuration file", e);
        }
      });

      this.mavenClient = MavenClientProvider.discoverProvider(DefaultToolingRuntimeClientBootstrap.class.getClassLoader())
          .createMavenClient(configuration.mavenConfiguration());
      this.toolingClassLoader = createClassLoader(configuration.toolingVersion());

      int nThreads = max(AVAILABLE_PROCESSORS * 2, 2);
      if (configuration.executorServiceConfiguration().isPresent()) {
        nThreads = configuration.executorServiceConfiguration().get().maxNumberOfThreads();
      }
      executorService = newFixedThreadPool(nThreads,
                                           runnable -> {
                                             Thread thread = new Thread(runnable, format("ToolingClient-%s-%s", threadNumber++,
                                                                                         identityHashCode(this)));
                                             thread.setUncaughtExceptionHandler((t, e) -> LOGGER.error("Internal error", e));
                                             return thread;
                                           });

      this.toolingRuntimeClientBuilderFactory =
          new ToolingRuntimeClientBuilderFactoryWrapper(configuration.muleVersion(),
                                                        toolingClassLoader,
                                                        executorService,
                                                        configuration.mavenConfiguration());
    } finally {
      resetSystemProperty(LOG4J2_DISABLE_JMX, currentLog4jDisableJmx);
      resetSystemProperty(LOG4J2_LOGGER_CONTEXT_FACTORY, currentLog4jLoggerContextFactory);
    }
  }

  private static void resetSystemProperty(String propertyName, String oldValue) {
    if (oldValue == null) {
      clearProperty(propertyName);
    } else {
      setProperty(propertyName, oldValue);
    }
  }

  private URLClassLoader createClassLoader(String toolingVersion) {
    try {
      if (LOGGER.isDebugEnabled()) {
        LOGGER.debug("Creating URL class loader for Tooling Runtime Client version: {}", toolingVersion);
      }

      Set<File> classpath = new LinkedHashSet<>();
      classpath.addAll(resolveDependency(newBuilder()
          .withGroupId(ORG_MULE_TOOLING)
          .withArtifactId(MULE_RUNTIME_TOOLING_CLIENT)
          .withVersion(toolingVersion)
          .build()));

      BundleDependency bundleDependency =
          mavenClient.resolveBundleDescriptor(new BundleDescriptor.Builder().setGroupId(ORG_MULE_TOOLING)
              .setArtifactId(MULE_RUNTIME_TOOLING_CLIENT)
              .setVersion(toolingVersion).setType(JAR).build());

      LOGGER.info("Resolved version of Tooling implementation: {}", bundleDependency);

      URL[] urls = new URL[] {};
      urls = addAll(urls, getBundleUrl(bundleDependency));
      urls = addAll(urls, classpath.stream()
          .filter(file -> getFileExtension(file.getAbsolutePath()).equals(JAR))
          .map(file -> toUrl(file))
          .toArray(URL[]::new));

      FilteringClassLoader jdkOnlyClassLoader = createJdkOnlyClassLoader();
      if (confDir != null) {
        urls = ArrayUtils.addAll(new URL[] {confDir.toURI().toURL()}, urls);
      }

      URLClassLoader classLoader =
          new ToolingClassLoader(configuration.muleVersion(), configuration.toolingVersion(), identityHashCode(this), urls,
                                 jdkOnlyClassLoader);

      if (LOGGER.isDebugEnabled()) {
        LOGGER.debug("Created Tooling Client class loader: {}", classLoader);
      }

      return classLoader;
    } catch (Exception e) {
      throw new ToolingException("Couldn't create the class loader for the Tooling Client", e);
    }
  }

  /**
   * @return {@link FilteringClassLoader} that only allows to resolve JDK classes, resources and declare as boot package
   * yourkit and takari file manager (this one in order to improve the performance for the concurrent local repository support).
   */
  // TODO MULE-11882 - Consolidate classloading isolation (including support for extend the boot packages list)
  private FilteringClassLoader createJdkOnlyClassLoader() {
    Set<String> packages = new HashSet(1024);
    Set<String> resources = new HashSet(1024);
    List<ExportedService> services = new ArrayList(128);
    JreExplorer.exploreJdk(packages, resources, services);
    ClassLoaderFilter classLoaderFilter = new ClassLoaderFilter(ImmutableSet.<String>builder().add("com.yourkit")
        .add("io.takari.filemanager").addAll(packages).build());
    return new FilteringClassLoader(classLoaderFilter, services);
  }

  private URL getBundleUrl(BundleDependency bundleDependency) {
    try {
      return bundleDependency.getBundleUri().toURL();
    } catch (MalformedURLException e) {
      throw new ToolingException("Error while getting bundleDependency URL", e);
    }
  }

  private List<File> resolveDependency(ArtifactDescriptor artifactDescriptor) {
    if (LOGGER.isDebugEnabled()) {
      LOGGER.debug("Resolving Tooling Runtime Client implementation libraries from: {}", artifactDescriptor);
    }

    List<BundleDependency> bundleDependencies = mavenClient
        .resolveBundleDescriptorDependencies(false, true,
                                             new BundleDescriptor.Builder()
                                                 .setGroupId(artifactDescriptor.getGroupId())
                                                 .setArtifactId(artifactDescriptor.getArtifactId())
                                                 .setVersion(artifactDescriptor.getVersion())
                                                 .setType(artifactDescriptor.getExtension())
                                                 .setClassifier(artifactDescriptor.getClassifier())
                                                 .build());

    return bundleDependencies.stream()
        .filter(bundleDependency -> bundleDependency.getBundleUri() != null)
        .map(bundleDependency -> new File(bundleDependency.getBundleUri()))
        .collect(toList());
  }

  private static URL toUrl(File file) {
    try {
      return file.toURI().toURL();
    } catch (MalformedURLException e) {
      throw new ToolingException("Couldn't get URL for file: " + file, e);
    }
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public String getMuleVersion() {
    return configuration.muleVersion();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public String getToolingVersion() {
    return configuration.toolingVersion();
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public ToolingRuntimeClientBuilderFactory getToolingRuntimeClientBuilderFactory() {
    if (toolingRuntimeClientBuilderFactory == null) {
      throw new IllegalStateException("Cannot be created a ToolingRuntimeClientBuilder once the bootstrap has been disposed");
    }
    return toolingRuntimeClientBuilderFactory;
  }

  /**
   * {@inheritDoc}
   */
  @Override
  public void dispose() {
    try {
      toolingRuntimeClientBuilderFactory.dispose();
    } finally {
      toolingRuntimeClientBuilderFactory = null;
    }

    executorService.shutdownNow();
    executorService = null;

    if (LOGGER.isDebugEnabled()) {
      LOGGER.debug("Closing Tooling Client class loader: {}", toolingClassLoader);
    }
    closeClassLoader(toolingClassLoader);
    toolingClassLoader = null;
    deleteQuietly(confDir);
  }

  private void closeClassLoader(URLClassLoader classLoader) {
    try {
      classLoader.close();
    } catch (IOException e) {
      // Do nothing
      if (LOGGER.isDebugEnabled()) {
        LOGGER.debug("Error while closing Tooling Client class loader", e);
      }
    }
  }

}
